/**
 * Get index with specific start index one by one. e.g.
 * min: 3, max: 9, start: 6
 *
 * Return index is:
 * [0]: 6
 * [1]: 7
 * [2]: 5
 * [3]: 8
 * [4]: 4
 * [5]: 9
 * [6]: 3
 */
export function getIndexByStartLoc(min, max, start, index) {
  var beforeCount = start - min;
  var afterCount = max - start;
  var balanceCount = Math.min(beforeCount, afterCount) * 2; // Balance

  if (index <= balanceCount) {
    var stepIndex = Math.floor(index / 2);

    if (index % 2) {
      return start + stepIndex + 1;
    }

    return start - stepIndex;
  } // One is out of range


  if (beforeCount > afterCount) {
    return start - (index - afterCount);
  }

  return start + (index - beforeCount);
}
/**
 * We assume that 2 list has only 1 item diff and others keeping the order.
 * So we can use dichotomy algorithm to find changed one.
 */

export function findListDiffIndex(originList, targetList, getKey) {
  if (originList.length === targetList.length) {
    return null;
  }

  var startIndex = 0;
  var endIndex = originList.length - 1;
  var midIndex = Math.floor((startIndex + endIndex) / 2);
  var keyCache = new Map();

  function getCacheKey(item) {
    if (!keyCache.has(item)) {
      keyCache.set(item, item !== undefined ? getKey(item) : {
        __EMPTY_ITEM__: true
      });
    }

    return keyCache.get(item);
  }

  while (startIndex !== midIndex || midIndex !== endIndex) {
    var originMidKey = getCacheKey(originList[midIndex]);
    var targetMidKey = getCacheKey(targetList[midIndex]);

    if (originMidKey === targetMidKey) {
      startIndex = midIndex;
    } else {
      endIndex = midIndex;
    } // Check if there only 2 index left


    if (startIndex === endIndex - 1) {
      return getCacheKey(originList[startIndex]) !== getCacheKey(targetList[startIndex]) ? startIndex : endIndex;
    }

    midIndex = Math.floor((startIndex + endIndex) / 2);
  }

  return midIndex;
}